from pbpstats.data_loader.data_nba.pbp.web import DataNbaPbpWebLoader


class DataNbaEnhancedPbpWebLoader(DataNbaPbpWebLoader):
    """
    A ``DataNbaEnhancedPbpWebLoader`` object should be instantiated and passed into ``DataNbaEnhancedPbpLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `data_<game_id>.json` in the `/pbp` subdirectory.
        If not provided response data will not be saved on disk.
    """

    def __init__(self, file_directory=None):
        self.file_directory = file_directory



"""
``DataLoaderFactory`` can be used to create data loader objects from the ``data_loader`` module.

The following code will create a data loader object for loading enhanced pbp from stats.nba.com.

.. code-block:: python

    from pbpstats.data_loader.factory import DataLoaderFactory

    data_loader = DataLoaderFactory()
    stats_enhanced_pbp_data_loader = data_loader.get_data_loader("stats_nba", "EnhancedPbp")
    print(stats_enhanced_pbp_data_loader[0])
    # prints "<class 'pbpstats.data_loader.stats_nba.enhanced_pbp_loader.StatsNbaEnhancedPbpLoader'>"
"""

from collections import defaultdict

import pbpstats.data_loader as data_loader


class DataLoaderFactory(object):
    """
    Class for factory of data loader classes. On initialization will load in all data loader classes in ``data_loader`` module
    """

    def __init__(self):
        self.loaders = defaultdict(lambda: defaultdict(lambda: []))
        self._load_data_loaders()

    def _load_data_loaders(self):
        """
        loads data loaders from data_loader package
        """
        loaders = dict(
            [
                (name, cls)
                for name, cls in data_loader.__dict__.items()
                if isinstance(cls, type)
            ]
        )
        for name, loader_cls in loaders.items():
            if hasattr(loader_cls, "resource"):
                file_source = loaders[name.replace("Loader", "FileLoader")]
                web_source = loaders[name.replace("Loader", "WebLoader")]
                loader = {
                    "loader": loader_cls,
                    "file_source": file_source,
                    "web_source": web_source,
                }
                self.loaders[loader_cls.resource][loader_cls.data_provider].append(
                    loader
                )

    def get_data_loader(self, data_provider, resource):
        """
        Gets data loader classes for given data provider and resource.

        :param str data_provider: Which data provider should data be loaded from. Options are 'stats_nba' or 'data_nba' or 'live'
        :param str resource: Name of class from resources directory

        :return: list of data loader classes
        :rtype: list
        """
        return self.loaders[resource][data_provider]



import json
import os
import requests
from pathlib import Path

from pbpstats import HEADERS, REQUEST_TIMEOUT, NBA_STRING, G_LEAGUE_STRING
from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader
from pbpstats.data_loader.stats_nba.summary.loader import StatsNbaSummaryLoader
from pbpstats.data_loader.stats_nba.summary.web import StatsNbaSummaryWebLoader


class StatsNbaShotsWebLoader(StatsNbaWebLoader):
    """
    A ``StatsNbaShotsWebLoader`` object should be instantiated and passed into ``StatsNbaShotsLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `stats_home_shots_<game_id>.json`
        and `stats_away_shots_<game_id>.json` in the `/game_details` subdirectory.
        If not provided response data will not be saved on disk.
    """

    def __init__(self, file_directory=None):
        self.file_directory = file_directory

    def load_data(self, game_id):
        self.game_id = game_id
        summary_source_loader = StatsNbaSummaryWebLoader(self.file_directory)
        summary = StatsNbaSummaryLoader(self.game_id, summary_source_loader)
        home_team_id = summary.items[0].home_team_id
        away_team_id = summary.items[0].visitor_team_id
        league_url_part = (
            f"{G_LEAGUE_STRING}.{NBA_STRING}"
            if self.league == G_LEAGUE_STRING
            else self.league
        )
        self.base_url = f"https://stats.{league_url_part}.com/stats/shotchartdetail"
        base_paramaters = {
            "GameID": self.game_id,
            "Season": self.season,
            "SeasonType": self.season_type,
            "PlayerID": 0,
            "Outcome": "",
            "Location": "",
            "Month": 0,
            "SeasonSegment": "",
            "DateFrom": "",
            "DateTo": "",
            "OpponentTeamID": 0,
            "VsConference": "",
            "VsDivision": "",
            "Position": "",
            "RookieYear": "",
            "GameSegment": "",
            "Period": 0,
            "LastNGames": 0,
            "ContextMeasure": "FG_PCT",
            "PlayerPosition": "",
            "LeagueID": self.game_id[0:2],
        }
        self.home_parameters = base_paramaters.copy()
        self.home_parameters["TeamID"] = home_team_id
        self.away_parameters = base_paramaters.copy()
        self.away_parameters["TeamID"] = away_team_id

        response = requests.get(
            self.base_url,
            self.home_parameters,
            headers=HEADERS,
            timeout=REQUEST_TIMEOUT,
        )
        if response.status_code == 200:
            self.home_source_data = response.json()
        else:
            response.raise_for_status()

        response = requests.get(
            self.base_url,
            self.away_parameters,
            headers=HEADERS,
            timeout=REQUEST_TIMEOUT,
        )
        if response.status_code == 200:
            self.away_source_data = response.json()
        else:
            response.raise_for_status()
        self._save_data_to_file()
        return self.home_source_data, self.away_source_data

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            home_file_path = f"{self.file_directory}/game_details/stats_home_shots_{self.game_id}.json"
            with open(home_file_path, "w") as outfile:
                json.dump(self.home_source_data, outfile)

            away_file_path = f"{self.file_directory}/game_details/stats_away_shots_{self.game_id}.json"
            with open(away_file_path, "w") as outfile:
                json.dump(self.away_source_data, outfile)



from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader


class StatsNbaScoreboardFileLoader(StatsNbaFileLoader):
    """
    A ``StatsNbaScoreboardFileLoader`` object should be instantiated and passed into ``StatsNbaScoreboardLoader`` when loading data from file

    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `stats_<league>_<date>.json` in the `/schedule` subdirectory.
    """

    def __init__(self, file_directory):
        self.file_directory = file_directory

    @check_file_directory
    def load_data(self, date, league_string):
        self.file_path = f'{self.file_directory}/schedule/stats_{league_string}_{date.replace("/", "_")}.json'
        return self._load_data_from_file()



from pbpstats.data_loader.live.pbp.file import LivePbpFileLoader


class LiveEnhancedPbpFileLoader(LivePbpFileLoader):
    """
    A ``LiveEnhancedPbpFileLoader`` object should be instantiated and passed into ``LiveEnhancedPbpLoader`` when loading data from file

    :param str game_id: NBA Stats Game Id
    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `live_<game_id>.json` in the `/pbp` subdirectory.
    """

    def __init__(self, file_directory):
        self.file_directory = file_directory



from pbpstats import (
    NBA_STRING,
    D_LEAGUE_STRING,
    WNBA_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)


class DataNbaLoaderBase(object):
    """
    Base Class for all data.nba.com data loaders

    This class should not be instantiated directly
    """

    @property
    def league(self):
        """
        Returns League for game id.

        First 2 in game id represent league - 00 for nba, 10 for wnba, 20 for g-league
        """
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return D_LEAGUE_STRING  # url uses dleague instead of gleague
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING

    @property
    def season(self):
        """
        Returns year in which season starts for game id

        4th and 5th characters in game id represent season year
        ex. for 2016-17 season 4th and 5th characters would be 16 and season should return 2016
        """
        if self.game_id[3] == "9":
            return "19" + self.game_id[3] + self.game_id[4]
        else:
            return "20" + self.game_id[3] + self.game_id[4]



from pbpstats.data_loader.stats_nba.enhanced_pbp.file import (
    StatsNbaEnhancedPbpFileLoader,
)


class StatsNbaPossessionFileLoader(object):
    """
    A ``StatsNbaPossessionFileLoader`` object should be instantiated and passed into ``StatsNbaPossessionLoader`` when loading data from file

    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `stats_<game_id>.json` in the `/pbp` subdirectory.
    """

    def __init__(self, file_directory):
        self.file_directory = file_directory
        self.enhanced_pbp_source_loader = StatsNbaEnhancedPbpFileLoader(file_directory)



"""
The ``Shots`` class has some basic properties for handling shot data
"""
from pbpstats.resources.base import Base


class Shots(Base):
    """
    Class for games items

    :param list items: list of
        :obj:`~pbpstats.resources.shots.stats_nba_shot.StatsNbaShot` items,
        typically from a shots data loader
    """

    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        """
        returns list of dicts with shots
        """
        return [item.data for item in self.items]



import json
import os

from pbpstats import NBA_STRING, G_LEAGUE_STRING

from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader


class StatsNbaGameFinderWebLoader(StatsNbaWebLoader):
    """
    A ``StatsNbaGameFinderWebLoader`` object should be instantiated and passed into ``StatsNbaGameFinderLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `stats_<league>_<season>_<season_type>.json` in the `/schedule` subdirectory.
        If not provided response data will not be saved on disk.
    """

    def __init__(self, file_directory=None):
        self.file_directory = file_directory

    def load_data(self, league, season, season_type):
        self.league_string = league
        self.season_string = season
        self.season_type_string = season_type
        league_url_part = (
            f"{G_LEAGUE_STRING}.{NBA_STRING}"
            if self.league_string == G_LEAGUE_STRING
            else self.league_string
        )
        self.base_url = f"https://stats.{league_url_part}.com/stats/leaguegamefinder"
        self.parameters = {
            "PlayerOrTeam": "T",
            "gtPTS": 1,
            "Season": self.season_string,
            "SeasonType": self.season_type_string,
            "LeagueID": self.league_id,
        }
        return self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.season_string.replace("-", "_")}_{self.season_type_string.replace(" ", "_")}.json'
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)



"""
``StatsNbaGameFinderLoader`` loads all games for a season and
creates :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem`
objects for each game

The following code will load data for the 2019-20 NBA Regular Season

.. code-block:: python

    from pbpstats.data_loader import StatsNbaGameFinderWebLoader, StatsNbaGameFinderLoader

    source_loader = StatsNbaGameFinderWebLoader("/data")
    game_finder_loader = StatsNbaGameFinderLoader("nba", "2019-20", "Regular Season", source_loader)
    print(game_finder_loader.items[0].data) # prints dict for first game
"""
from pbpstats.data_loader.stats_nba.base import StatsNbaLoaderBase
from pbpstats.resources.games.stats_nba_game_item import StatsNbaGameItem


class StatsNbaGameFinderLoader(StatsNbaLoaderBase):
    """
    Loads stats.nba.com source data for season.
    Games are stored in items attribute
    as :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem` objects

    :param str league: Options are 'nba', 'wnba' or 'gleague'
    :param str season: Formatted as 2019-20 for NBA and G-League, 2019 of WNBA.
    :param str season_type: Options are 'Regular Season' or 'Playoffs' or 'Play In'
    :param source_loader: :obj:`~pbpstats.data_loader.stats_nba.game_finder.file.StatsNbaGameFinderFileLoader` or :obj:`~pbpstats.data_loader.stats_nba.game_finder.web.StatsNbaGameFinderWebLoader` object
    """

    data_provider = "stats_nba"
    resource = "Games"
    parent_object = "Season"

    def __init__(self, league, season, season_type, source_loader):
        self.league_string = league
        self.season_string = season
        self.season_type_string = season_type
        self.source_data = source_loader.load_data(league, season, season_type)
        self._make_game_data_items()

    def _make_game_data_items(self):
        self.items = []
        sorted_games = sorted(self.data, key=lambda k: k["GAME_ID"])
        for team1, team2 in zip(sorted_games[0::2], sorted_games[1::2]):
            item = self._make_single_dict_for_game(team1, team2)
            self.items.append(StatsNbaGameItem(item))

    @staticmethod
    def _make_single_dict_for_game(team1, team2):
        """
        MATCHUP is vistor team @ home team for visitor
        MATCHUP is home team vs visitor team for home
        """
        item = {
            "GAME_ID": team1["GAME_ID"],
            "GAME_DATE_EST": team1["GAME_DATE"],
            "GAME_STATUS_TEXT": "Final",
            "HOME_TEAM_ID": team1["TEAM_ID"]
            if "@" not in team1["MATCHUP"]
            else team2["TEAM_ID"],
            "VISITOR_TEAM_ID": team1["TEAM_ID"]
            if "@" in team1["MATCHUP"]
            else team2["TEAM_ID"],
        }
        return item



import json
import os

from pbpstats.data_loader.live.web_loader import LiveWebLoader


class LivePbpWebLoader(LiveWebLoader):
    """
    A ``LivePbpWebLoader`` object should be instantiated and passed into ``LivePbpLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `live_<game_id>.json` in the `/pbp` subdirectory.
        If not provided response data will not be saved on disk.
    """

    def __init__(self, file_directory=None):
        self.file_directory = file_directory

    def load_data(self, game_id):
        self.game_id = game_id
        league = self.league.upper()
        self.url = f"https://nba-prod-us-east-1-mediaops-stats.s3.amazonaws.com/{league}/liveData/playbyplay/playbyplay_{self.game_id}.json"
        return self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f"{self.file_directory}/pbp/live_{self.game_id}.json"
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)



"""
``StatsNbaShotsLoader`` loads shot data for a game and
creates :obj:`~pbpstats.resources.shots.stats_nba_shot.StatsNbaShot`
objects for all shots

The following code will load shot data for game id "0021900001" from
a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import StatsNbaShotsFileLoader, StatsNbaShotsLoader

    source_loader = StatsNbaShotsFileLoader("/data")
    shot_loader = StatsNbaShotsLoader("0021900001", source_loader)
    print(shot_loader.items[0].data) # prints dict with data for one shot from game
"""
from pbpstats.data_loader.stats_nba.base import StatsNbaLoaderBase
from pbpstats.resources.shots.stats_nba_shot import StatsNbaShot


class StatsNbaShotsLoader(StatsNbaLoaderBase):
    """
    Loads stats.nba.com source shot data for game.
    Shots are stored in items attribute
    as :obj:`~pbpstats.resources.shots.stats_nba_shot.StatsNbaShot` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.stats_nba.shots.file.StatsNbaShotsFileLoader` or :obj:`~pbpstats.data_loader.stats_nba.shots.web.StatsNbaShotsWebLoader` object
    """

    data_provider = "stats_nba"
    resource = "Shots"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.game_id = game_id
        self.home_source_data, self.away_source_data = source_loader.load_data(
            self.game_id
        )
        self._make_shot_items()

    def _make_shot_items(self):
        self.items = [StatsNbaShot(item) for item in self.data]

    def make_list_of_dicts(self, results_set_index=0):
        """
        Creates list of dicts from home and away source data

        :param int results_set_index: Index results are in. Default is 0
        :returns: list of dicts with shot data for all shots
        """
        headers = self.home_source_data["resultSets"][results_set_index]["headers"]
        home_rows = self.home_source_data["resultSets"][results_set_index]["rowSet"]
        home_deduped_rows = self.dedupe_events_row_set(home_rows)
        away_rows = self.away_source_data["resultSets"][results_set_index]["rowSet"]
        away_deduped_rows = self.dedupe_events_row_set(away_rows)
        return [dict(zip(headers, row)) for row in home_deduped_rows] + [
            dict(zip(headers, row)) for row in away_deduped_rows
        ]



"""
``EnhancedPbpItem`` is an abstract base class for all enhanced pbp event types
"""
import abc

import pbpstats
from pbpstats.resources.enhanced_pbp import FieldGoal, Foul, FreeThrow, Rebound


class EnhancedPbpItem(metaclass=abc.ABCMeta):
    def __repr__(self):
        return f"<{type(self).__name__} GameId: {self.game_id}, Description: {self.description}, Time: {self.clock}, EventNum: {self.event_num}>"

    @abc.abstractproperty
    def is_possession_ending_event(self):
        """
        returns True if event ends a possession, False otherwise
        """
        pass

    @abc.abstractproperty
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        pass

    @abc.abstractmethod
    def get_offense_team_id(self):
        """
        returns team id for team on offense for event
        """
        pass

    @abc.abstractproperty
    def seconds_remaining(self):
        """
        returns seconds remaining in period as a ``float``
        """
        pass

    @property
    def base_stats(self):
        """
        returns list of dicts with all seconds played and possession count stats for event
        """
        return (
            self._get_seconds_played_stats_items()
            + self._get_possessions_played_stats_items()
        )

    def get_all_events_at_current_time(self):
        """
        returns list of all events that take place as the same time as the current event
        """
        events = [self]
        # going backwards
        event = self
        while event is not None and self.seconds_remaining == event.seconds_remaining:
            if event != self:
                events.append(event)
            event = event.previous_event
        # going forwards
        event = self
        while event is not None and self.seconds_remaining == event.seconds_remaining:
            if event != self:
                events.append(event)
            event = event.next_event
        return sorted(events, key=lambda k: k.order)

    @property
    def current_players(self):
        """
        returns dict with list of player ids for each team
        with players on the floor for current event

        For all non subsitution events current players are just
        the same as previous event

        This gets overwritten in :obj:`~pbpstats.resources.enhanced_pbp.substitution.Substitution`
        since those are the only event types where players are not the same as the previous event
        """
        return self.previous_event.current_players

    @property
    def score_margin(self):
        """
        returns the score margin from perspective of offense team before the event took place
        """
        if self.previous_event is None:
            score = self.score
        else:
            score = self.previous_event.score
        offense_team_id = self.get_offense_team_id()
        offense_points = score[offense_team_id]
        defense_points = 0
        for team_id, points in score.items():
            if team_id != offense_team_id:
                defense_points = points
        return offense_points - defense_points

    @property
    def lineup_ids(self):
        """
        returns dict with lineup ids for each team for current event.
        Lineup ids are hyphen separated sorted player id strings.
        """
        lineup_ids = {}
        for team_id, team_players in self.current_players.items():
            players = [str(player_id) for player_id in team_players]
            sorted_player_ids = sorted(players)
            lineup_id = "-".join(sorted_player_ids)
            lineup_ids[team_id] = lineup_id
        return lineup_ids

    @property
    def seconds_since_previous_event(self):
        """
        returns the number of seconds that have elapsed since the previous event
        """
        if self.previous_event is None:
            return 0
        if self.seconds_remaining == 720:
            # so that subs between periods for live don't have negative seconds
            return 0
        if self.seconds_remaining == 300 and self.period > 4:
            # so that subs between periods for live don't have negative seconds
            return 0
        return self.previous_event.seconds_remaining - self.seconds_remaining

    def is_second_chance_event(self):
        """
        returns True if the event takes place after an offensive rebound
        on the current possession, False otherwise
        """
        event = self.previous_event
        if isinstance(event, Rebound) and event.is_real_rebound and event.oreb:
            return True
        while not (event is None or event.is_possession_ending_event):
            if isinstance(event, Rebound) and event.is_real_rebound and event.oreb:
                return True
            event = event.previous_event
        return False

    def is_penalty_event(self):
        """
        returns True if the team on offense is in the penalty, False otherwise
        """
        if hasattr(self, "fouls_to_give"):
            team_ids = list(self.current_players.keys())
            offense_team_id = self.get_offense_team_id()
            defense_team_id = (
                team_ids[0] if offense_team_id == team_ids[1] else team_ids[1]
            )
            if self.fouls_to_give[defense_team_id] == 0:
                if isinstance(self, (Foul, FreeThrow, Rebound)):
                    # if foul or free throw or rebound on a missed ft
                    # check foul event and should return false is foul
                    # was shooting foul and team had a foul to give
                    if isinstance(self, Foul):
                        foul_event = self
                    elif isinstance(self, FreeThrow):
                        foul_event = self.foul_that_led_to_ft
                    else:
                        # if rebound is on missed ft, also need to look at foul that led to FT
                        if not self.oreb and isinstance(self.missed_shot, FreeThrow):
                            foul_event = self.missed_shot.foul_that_led_to_ft
                        else:
                            return True
                    if foul_event is None:
                        return True
                    fouls_to_give_prior_to_foul = (
                        foul_event.previous_event.fouls_to_give[defense_team_id]
                    )
                    if fouls_to_give_prior_to_foul > 0:
                        return False
                return True
        return False

    @property
    def count_as_possession(self):
        """
        returns True if event is possession changing event
        that should count as a real possession, False otherwise.

        In order to not include possessions which a very low probability of scoring in possession counts,
        possession won't be counted as a possession if it starts with <= 2 seconds left
        and no points are scored before period ends
        """
        if self.is_possession_ending_event:
            if self.seconds_remaining > 2:
                return True
            # check when previous possession ended
            prev_event = self.previous_event
            while prev_event is not None and not prev_event.is_possession_ending_event:
                prev_event = prev_event.previous_event
            if prev_event is None or prev_event.seconds_remaining > 2:
                return True
            # possession starts in final 2 seconds
            # return True if there is a FT or FGM between now and end of period
            next_event = prev_event.next_event
            while next_event is not None:
                if isinstance(next_event, FreeThrow) or (
                    isinstance(next_event, FieldGoal) and next_event.is_made
                ):
                    return True
                next_event = next_event.next_event
        return False

    def _get_seconds_played_stats_items(self):
        """
        makes event stats items for:
        - seconds played
        - seconds played for number of fouls
        - second chance seconds played
        - penalty seconds played
        """
        stat_items = []
        team_ids = list(self.current_players.keys())
        offense_team_id = self.get_offense_team_id()
        is_penalty_event = self.is_penalty_event()
        is_second_chance_event = self.is_second_chance_event()
        if self.seconds_since_previous_event != 0:
            for team_id, players in self.previous_event.current_players.items():
                seconds_stat_key = (
                    pbpstats.SECONDS_PLAYED_OFFENSE_STRING
                    if team_id == offense_team_id
                    else pbpstats.SECONDS_PLAYED_DEFENSE_STRING
                )
                opponent_team_id = (
                    team_ids[0] if team_id == team_ids[1] else team_ids[1]
                )
                previous_poss_lineup_ids = self.previous_event.lineup_ids
                for player_id in players:
                    keys_to_add = [seconds_stat_key]
                    player_fouls = self.previous_event.player_game_fouls[player_id]
                    period = self.period if self.period <= 4 else "OT"
                    foul_tracking_seconds_stat_key = (
                        f"Period{period}Fouls{player_fouls}{seconds_stat_key}"
                    )
                    keys_to_add.append(foul_tracking_seconds_stat_key)
                    if is_second_chance_event:
                        seconds_chance_seconds_stat_key = (
                            f"{pbpstats.SECOND_CHANCE_STRING}{seconds_stat_key}"
                        )
                        keys_to_add.append(seconds_chance_seconds_stat_key)
                    if is_penalty_event:
                        penalty_seconds_stat_key = (
                            f"{pbpstats.PENALTY_STRING}{seconds_stat_key}"
                        )
                        keys_to_add.append(penalty_seconds_stat_key)
                    for stat_key in keys_to_add:
                        stat_item = {
                            "player_id": player_id,
                            "team_id": team_id,
                            "opponent_team_id": opponent_team_id,
                            "lineup_id": previous_poss_lineup_ids[team_id],
                            "opponent_lineup_id": previous_poss_lineup_ids[
                                opponent_team_id
                            ],
                            "stat_key": stat_key,
                            "stat_value": self.seconds_since_previous_event,
                        }
                        stat_items.append(stat_item)
        return stat_items

    def _get_possessions_played_stats_items(self):
        """
        makes event stats items for:
        - possessions played
        - second chance possessions played
        - penalty possessions played
        """
        stat_items = []
        team_ids = list(self.current_players.keys())
        offense_team_id = self.get_offense_team_id()
        is_penalty_event = self.is_penalty_event()
        is_second_chance_event = self.is_second_chance_event()
        if self.count_as_possession:
            if isinstance(self, FreeThrow):
                current_players = self.event_for_efficiency_stats.current_players
                lineup_ids = self.event_for_efficiency_stats.lineup_ids
            else:
                current_players = self.current_players
                lineup_ids = self.lineup_ids
            for team_id, players in current_players.items():
                possessions_stat_key = (
                    pbpstats.OFFENSIVE_POSSESSION_STRING
                    if team_id == offense_team_id
                    else pbpstats.DEFENSIVE_POSSESSION_STRING
                )
                opponent_team_id = (
                    team_ids[0] if team_id == team_ids[1] else team_ids[1]
                )
                for player_id in players:
                    keys_to_add = [possessions_stat_key]
                    if is_second_chance_event:
                        seconds_chance_possessions_stat_key = (
                            f"{pbpstats.SECOND_CHANCE_STRING}{possessions_stat_key}"
                        )
                        keys_to_add.append(seconds_chance_possessions_stat_key)
                    if is_penalty_event:
                        penalty_possessions_stat_key = (
                            f"{pbpstats.PENALTY_STRING}{possessions_stat_key}"
                        )
                        keys_to_add.append(penalty_possessions_stat_key)
                    for stat_key in keys_to_add:
                        stat_item = {
                            "player_id": player_id,
                            "team_id": team_id,
                            "opponent_team_id": opponent_team_id,
                            "lineup_id": lineup_ids[team_id],
                            "opponent_lineup_id": lineup_ids[opponent_team_id],
                            "stat_key": stat_key,
                            "stat_value": 1,
                        }
                        stat_items.append(stat_item)

        return stat_items



"""
``StatsNbaSummaryLoader`` loads summary data for a game and
creates :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem`
objects for game

The following code will load summary data for game id "0021900001" from
a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader.stats_nba.summary.file import StatsNbaSummaryFileLoader
    from pbpstats.data_loader.stats_nba.summary.loader import StatsNbaSummaryLoader

    source_loader = StatsNbaSummaryFileLoader("/data")
    summary_loader = StatsNbaSummaryLoader("0021900001", source_loader)
    print(summary_loader.items[0].data) # prints game summary dict for game
"""
from pbpstats.data_loader.stats_nba.base import StatsNbaLoaderBase
from pbpstats.resources.games.stats_nba_game_item import StatsNbaGameItem


class StatsNbaSummaryLoader(StatsNbaLoaderBase):
    """
    Loads stats.nba.com source summary data for game.
    Summary data is stored in items attribute
    as :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.stats_nba.summary.file.StatsNbaSummaryFileLoader` or :obj:`~pbpstats.data_loader.stats_nba.summary.web.StatsNbaSummaryWebLoader` object
    """

    data_provider = "stats_nba"
    resource = "Games"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.game_id = game_id
        self.source_data = source_loader.load_data(self.game_id)
        self._make_summary_items()

    def _make_summary_items(self):
        self.items = [StatsNbaGameItem(item) for item in self.data]



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import Violation


class LiveViolation(Violation, LiveEnhancedPbpItem):
    """
    Class for violation events
    """

    action_type = "violation"

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def is_delay_of_game(self):
        return self.stripped_sub_type == "delayofgame"

    @property
    def is_goaltend_violation(self):
        return self.stripped_sub_type == "defensivegoaltending"

    @property
    def is_lane_violation(self):
        return self.stripped_sub_type == "lane"

    @property
    def is_jumpball_violation(self):
        return self.stripped_sub_type == "jumpball"

    @property
    def is_kicked_ball_violation(self):
        return self.stripped_sub_type == "kickedball"

    @property
    def is_double_lane_violation(self):
        return self.stripped_sub_type == "doublelane"



import abc


class Substitution(object):
    """
    Class for Substitution events
    """

    @abc.abstractproperty
    def outgoing_player_id(self):
        pass

    @abc.abstractproperty
    def incoming_player_id(self):
        pass

    @property
    def current_players(self):
        """
        returns dict with list of player ids for each team
        with players on the floor following the sub
        """
        players = self.previous_event.current_players.copy()
        players[self.team_id] = [
            self.incoming_player_id if player == self.outgoing_player_id else player
            for player in players[self.team_id]
        ]
        return players

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        return self.base_stats



"""
The ``Possessions`` class has some basic properties for aggregating possession stats
"""
from itertools import groupby
from operator import itemgetter

from pbpstats import KEYS_OFF_BY_FACTOR_OF_5_WHEN_AGGREGATING_FOR_TEAM_AND_LINEUPS
from pbpstats.resources.base import Base


class Possessions(Base):
    """
    Class for possession items

    :param list items: list of
        :obj:`~pbpstats.resources.possessions.possession.Possession` items,
        typically from a possession data loader
    """

    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        """
        returns possessions dict
        """
        return self.__dict__

    def _aggregate_event_stats(self, *args):
        stats = [
            event_stat
            for item in self.items
            for event in item.events
            for event_stat in event.event_stats
        ]
        grouper = itemgetter(*args)
        results = []
        for key, grp in groupby(sorted(stats, key=grouper), grouper):
            temp_dict = dict(zip([*args], key))
            value = sum(item["stat_value"] for item in grp)
            if (
                temp_dict["stat_key"]
                in KEYS_OFF_BY_FACTOR_OF_5_WHEN_AGGREGATING_FOR_TEAM_AND_LINEUPS
                and "player_id" not in args
            ):
                # since stat keys are summed up from player stats
                # team and lineup stats will need some stats to be divided by 5
                value = value / 5
            temp_dict["stat_value"] = (
                value if isinstance(value, int) else round(value, 1)
            )
            results.append(temp_dict)
        return results

    @property
    def team_stats(self):
        """
        returns list of dicts with aggregated stats by team
        """
        return self._aggregate_event_stats("team_id", "stat_key")

    @property
    def opponent_stats(self):
        """
        returns list of dicts with aggregated stats by opponent
        """
        return self._aggregate_event_stats("opponent_team_id", "stat_key")

    @property
    def player_stats(self):
        """
        returns list of dicts with aggregated stats by player
        """
        return self._aggregate_event_stats("player_id", "team_id", "stat_key")

    @property
    def lineup_stats(self):
        """
        returns list of dicts with aggregated stats by lineup
        """
        return self._aggregate_event_stats("lineup_id", "team_id", "stat_key")

    @property
    def lineup_opponent_stats(self):
        """
        returns list of dicts with aggregated stats by lineup opponent
        """
        return self._aggregate_event_stats(
            "opponent_lineup_id", "opponent_team_id", "stat_key"
        )



"""
``StatsNbaBoxscoreLoader`` loads boxscore data for a game and
creates :obj:`~pbpstats.resources.boxscore.stats_nba_boxscore_item.StatsNbaBoxscoreItem`
objects for each player and team

The following code will load boxscore data for game id "0021900001" from
a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import StatsNbaBoxscoreFileLoader, StatsNbaBoxscoreLoader

    source_loader = StatsNbaBoxscoreFileLoader("/data")
    boxscore_loader = StatsNbaBoxscoreLoader("0021900001", source_loader)
    print(boxscore_loader.items[0].data) # prints dict with a player's boxscore data for game
"""
from pbpstats.data_loader.stats_nba.base import StatsNbaLoaderBase
from pbpstats.resources.boxscore.stats_nba_boxscore_item import StatsNbaBoxscoreItem


class StatsNbaBoxscoreLoader(StatsNbaLoaderBase):
    """
    Loads stats.nba.com source boxscore data for game.
    Team/Player data is stored in items attribute as :obj:`~pbpstats.resources.boxscore.stats_nba_boxscore_item.StatsNbaBoxscoreItem` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.stats_nba.boxscore.file.StatsNbaBoxscoreFileLoader` or :obj:`~pbpstats.data_loader.stats_nba.boxscore.web.StatsNbaBoxscoreWebLoader` object
    """

    data_provider = "stats_nba"
    resource = "Boxscore"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.game_id = game_id
        self.source_data = source_loader.load_data(self.game_id)
        self._make_boxscore_items()

    def _make_boxscore_items(self):
        self.items = [StatsNbaBoxscoreItem(item) for item in self.data]
        self.items += [
            StatsNbaBoxscoreItem(item)
            for item in self.make_list_of_dicts(results_set_index=1)
        ]



from pbpstats.resources.enhanced_pbp.live.ejection import LiveEjection
from pbpstats.resources.enhanced_pbp.live.end_of_period import LiveEndOfPeriod
from pbpstats.resources.enhanced_pbp.live.field_goal import LiveFieldGoal
from pbpstats.resources.enhanced_pbp.live.foul import LiveFoul
from pbpstats.resources.enhanced_pbp.live.free_throw import LiveFreeThrow
from pbpstats.resources.enhanced_pbp.live.jump_ball import LiveJumpBall
from pbpstats.resources.enhanced_pbp.live.rebound import LiveRebound
from pbpstats.resources.enhanced_pbp.live.replay import LiveReplay
from pbpstats.resources.enhanced_pbp.live.start_of_period import LiveStartOfPeriod
from pbpstats.resources.enhanced_pbp.live.substitution import LiveSubstitution
from pbpstats.resources.enhanced_pbp.live.timeout import LiveTimeout
from pbpstats.resources.enhanced_pbp.live.turnover import LiveTurnover
from pbpstats.resources.enhanced_pbp.live.violation import LiveViolation

__all__ = [
    "LiveEjection",
    "LiveEndOfPeriod",
    "LiveFieldGoal",
    "LiveFoul",
    "LiveFreeThrow",
    "LiveJumpBall",
    "LiveRebound",
    "LiveReplay",
    "LiveStartOfPeriod",
    "LiveSubstitution",
    "LiveTimeout",
    "LiveTurnover",
    "LiveViolation",
]



"""
``StatsNbaScoreboardLoader`` loads all games for a date and
creates :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem`
objects for each game

The following code will load data for 02/03/2020

.. code-block:: python

    from pbpstats.data_loader import StatsNbaScoreboardFileLoader, StatsNbaScoreboardLoader

    source_loader = StatsNbaScoreboardFileLoader("/data")
    game_finder_loader = StatsNbaScoreboardLoader("02/03/2020", "nba", source_loader)
    print(game_finder_loader.items[0].data) # prints dict for first game
"""
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader
from pbpstats.resources.games.stats_nba_game_item import StatsNbaGameItem


class StatsNbaScoreboardLoader(StatsNbaFileLoader, StatsNbaWebLoader):
    """
    Loads stats.nba.com source data for date.
    Games are stored in items attribute
    as :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem` objects

    :param str date: Formatted as MM/DD/YYYY
    :param str league_string: Options are 'nba', 'wnba' or 'gleague'
    :param source_loader: :obj:`~pbpstats.data_loader.stats_nba.scoreboard.file.StatsNbaScoreboardFileLoader` or :obj:`~pbpstats.data_loader.stats_nba.scoreboard.web.StatsNbaScoreboardWebLoader` object
    """

    data_provider = "stats_nba"
    resource = "Games"
    parent_object = "Day"

    def __init__(self, date, league_string, source_loader):
        self.date = date
        self.league_string = league_string
        self.source_data = source_loader.load_data(date, league_string)
        self._make_scoreboard_items()

    def _make_scoreboard_items(self):
        self.items = [StatsNbaGameItem(item) for item in self.data]



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Foul


class StatsFoul(Foul, StatsEnhancedPbpItem):
    """
    Class for foul events
    """

    event_type = 6

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def number_of_fta_for_foul(self):
        """
        returns the number of free throws resulting from the foul
        """
        clock = self.clock
        event = self
        while (
            event is not None
            and event.clock == clock
            and not (
                hasattr(event, "is_first_ft")
                and not event.is_technical_ft
                and self.team_id != event.team_id
            )
        ):
            event = event.next_event

        if (
            event is not None
            and hasattr(event, "is_first_ft")
            and not event.is_technical_ft
            and event.clock == clock
            and (not hasattr(self, "player3_id") or self.player3_id == event.player1_id)
        ):
            # player3 id check is to make sure player who got fouled is player shooting free throws, prior to 2005-06 because foul drawning player isn't in pbp
            if "of 1" in event.description:
                return 1
            elif "of 2" in event.description:
                return 2
            elif "of 3" in event.description:
                return 3

        # if we haven't found ft yet, try going backwards
        event = self
        while (
            event is not None
            and event.clock == clock
            and not (
                hasattr(event, "is_first_ft")
                and not event.is_technical_ft
                and self.team_id != event.team_id
            )
        ):
            event = event.previous_event

        if (
            event is not None
            and hasattr(event, "is_first_ft")
            and not event.is_technical_ft
            and event.clock == clock
            and (not hasattr(self, "player3_id") or self.player3_id == event.player1_id)
        ):
            # player3 id check is to make sure player who got fouled is player shooting free throws, prior to 2005-06 because foul drawning player isn't in pbp
            if "of 1" in event.description:
                return 1
            elif "of 2" in event.description:
                return 2
            elif "of 3" in event.description:
                return 3
        return None

    @property
    def is_personal_foul(self):
        return self.event_action_type == 1

    @property
    def is_shooting_foul(self):
        return self.event_action_type == 2

    @property
    def is_loose_ball_foul(self):
        return self.event_action_type == 3

    @property
    def is_offensive_foul(self):
        return self.event_action_type == 4

    @property
    def is_inbound_foul(self):
        return self.event_action_type == 5

    @property
    def is_away_from_play_foul(self):
        return self.event_action_type == 6

    @property
    def is_clear_path_foul(self):
        return self.event_action_type == 9

    @property
    def is_double_foul(self):
        return self.event_action_type == 10

    @property
    def is_technical(self):
        return self.event_action_type in [11, 12, 13, 18, 19, 25, 30]

    @property
    def is_flagrant1(self):
        return self.event_action_type == 14

    @property
    def is_flagrant2(self):
        return self.event_action_type == 15

    @property
    def is_double_technical(self):
        return self.event_action_type == 16

    @property
    def is_defensive_3_seconds(self):
        return self.event_action_type == 17

    @property
    def is_delay_of_game(self):
        return self.event_action_type == 18

    @property
    def is_charge(self):
        return self.event_action_type == 26

    @property
    def is_personal_block_foul(self):
        return self.event_action_type == 27

    @property
    def is_personal_take_foul(self):
        return self.event_action_type == 28

    @property
    def is_shooting_block_foul(self):
        return self.event_action_type == 29



import os
import json
from collections import defaultdict

from pbpstats import NBA_STRING
from pbpstats.overrides import IntDecoder
from pbpstats.resources.enhanced_pbp import FieldGoal, Foul, FreeThrow, StartOfPeriod


class NbaEnhancedPbpLoader(object):
    """
    Class for shared methods between :obj:`~pbpstats.data_loader.data_nba.enhanced_pbp_loader.DataNbaEnhancedPbpLoader`
    and :obj:`~pbpstats.data_loader.stats_nba.enhanced_pbp_loader.StatsNbaEnhancedPbpLoader`

    Both :obj:`~pbpstats.data_loader.data_nba.enhanced_pbp_loader.DataNbaEnhancedPbpLoader`
    and :obj:`~pbpstats.data_loader.stats_nba.enhanced_pbp_loader.StatsNbaEnhancedPbpLoader` should inherit from this class

    This class should not be instantiated directly
    """

    def _add_extra_attrs_to_all_events(self):
        """
        adds fouls to give, player fouls, score, next event and previous event to each event
        """
        self.start_period_indices = []
        self._load_possession_changing_event_overrides()
        game_id = self.game_id if self.league == NBA_STRING else int(self.game_id)
        change_override_event_nums = self.possession_changing_event_overrides.get(
            game_id, []
        )
        non_change_override_event_nums = (
            self.non_possession_changing_event_overrides.get(game_id, [])
        )
        player_game_fouls = defaultdict(int)
        fouls_to_give = defaultdict(lambda: 4)
        score = defaultdict(int)
        for i, event in enumerate(self.items):
            if i == 0 and i == len(self.items) - 1:
                event.previous_event = None
                event.next_event = None
            elif isinstance(event, StartOfPeriod) or i == 0:
                event.previous_event = None
                event.next_event = self.items[i + 1]
                self.start_period_indices.append(i)
                if event.period <= 4:
                    fouls_to_give = defaultdict(lambda: 4)
                else:
                    fouls_to_give = defaultdict(lambda: 3)
            elif i == len(self.items) - 1 or event.period != self.items[i + 1].period:
                event.previous_event = self.items[i - 1]
                event.next_event = None
            else:
                event.previous_event = self.items[i - 1]
                event.next_event = self.items[i + 1]

            if event.seconds_remaining <= 120:
                if len(fouls_to_give.keys()) == 0:
                    # neither team has fouled yet in the period
                    fouls_to_give = defaultdict(lambda: 1)
                elif len(fouls_to_give.keys()) == 1:
                    # only one team has fouled - other team id key is not in defaultdict
                    team_id = list(fouls_to_give.keys())[0]
                    team_fouls_to_give = min(fouls_to_give[team_id], 1)
                    fouls_to_give = defaultdict(lambda: 1)
                    fouls_to_give[team_id] = team_fouls_to_give
                else:
                    for team_id in fouls_to_give.keys():
                        fouls_to_give[team_id] = min(fouls_to_give[team_id], 1)
            if isinstance(event, Foul):
                if event.counts_towards_penalty and fouls_to_give[event.team_id] > 0:
                    fouls_to_give[event.team_id] -= 1
                if event.counts_as_personal_foul:
                    player_game_fouls[event.player1_id] += 1
            if isinstance(event, (FieldGoal, FreeThrow)) and event.is_made:
                score[event.team_id] += event.shot_value

            event.fouls_to_give = fouls_to_give.copy()
            event.player_game_fouls = player_game_fouls.copy()
            event.score = score.copy()
            event.possession_changing_override = (
                event.event_num in change_override_event_nums
            )
            event.non_possession_changing_override = (
                event.event_num in non_change_override_event_nums
            )

        # these need next and previous event to be added to all events
        self._set_period_start_items()

    def _set_period_start_items(self):
        """
        sets team starting period with the ball and period starters for each team
        """
        for i in self.start_period_indices:
            team_id = self.items[i].get_team_starting_with_ball()
            self.items[i].team_starting_with_ball = team_id
            period_starters = self.items[i].get_period_starters(
                file_directory=self.file_directory
            )
            self.items[i].period_starters = period_starters

    def _load_possession_changing_event_overrides(self):
        """
        loads overrides for possession or non possession changing events
        """
        if self.file_directory is not None:
            possession_changing_event_overrides_file_path = f"{self.file_directory}/overrides/possession_change_event_overrides.json"
            if os.path.isfile(possession_changing_event_overrides_file_path):
                with open(possession_changing_event_overrides_file_path) as f:
                    # issues with pbp - force these events to be possession changing events
                    # {GameId: [EventNum]}
                    self.possession_changing_event_overrides = json.loads(
                        f.read(), cls=IntDecoder
                    )
            else:
                self.possession_changing_event_overrides = {}

            non_possession_changing_event_overrides_file_path = f"{self.file_directory}/overrides/non_possession_changing_event_overrides.json"
            if os.path.isfile(non_possession_changing_event_overrides_file_path):
                with open(non_possession_changing_event_overrides_file_path) as f:
                    # issues with pbp - force these events to be not possession changing events
                    # {GameId: [EventNum]}
                    self.non_possession_changing_event_overrides = json.loads(
                        f.read(), cls=IntDecoder
                    )
            else:
                self.non_possession_changing_event_overrides = {}
        else:
            self.possession_changing_event_overrides = {}
            self.non_possession_changing_event_overrides = {}



"""
``LivePossessionLoader`` loads possession data for a game and creates :obj:`~pbpstats.resources.possessions.possession.Possession` objects for each possession

The following code will load possession data for game id "0021900001" from a pbp file located in the ``/pbp`` subdirectory of the ``/data`` directory

.. code-block:: python

    from pbpstats.data_loader import LivePossessionFileLoader, LivePossessionLoader

    source_loader = LivePossessionFileLoader("/data")
    pbp_loader = LivePossessionLoader("0021900001", source_loader)
    print(pbp_loader.items[0].data)  # prints dict with the first event of the game
"""
from pbpstats.data_loader.live.enhanced_pbp.loader import LiveEnhancedPbpLoader
from pbpstats.data_loader.nba_possession_loader import NbaPossessionLoader
from pbpstats.resources.possessions.possession import Possession


class LivePossessionLoader(NbaPossessionLoader):
    """
    Loads live data source possession data for game.
    Possessions are stored in items attribute as :obj:`~pbpstats.resources.possessions.possession.Possession` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.live.possessions.file.LivePossessionFileLoader` or :obj:`~pbpstats.data_loader.live.possessions.web.LivePossessionWebLoader` object
    """

    data_provider = "live"
    resource = "Possessions"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.game_id = game_id
        self.file_directory = source_loader.file_directory
        pbp_events = LiveEnhancedPbpLoader(
            game_id, source_loader.enhanced_pbp_source_loader
        )
        self.events = pbp_events.items
        events_by_possession = self._split_events_by_possession()
        self.items = [
            Possession(possession_events) for possession_events in events_by_possession
        ]
        self._add_extra_attrs_to_all_possessions()



from pbpstats.resources.enhanced_pbp.data_nba.ejection import DataEjection
from pbpstats.resources.enhanced_pbp.data_nba.end_of_period import DataEndOfPeriod
from pbpstats.resources.enhanced_pbp.data_nba.field_goal import DataFieldGoal
from pbpstats.resources.enhanced_pbp.data_nba.foul import DataFoul
from pbpstats.resources.enhanced_pbp.data_nba.free_throw import DataFreeThrow
from pbpstats.resources.enhanced_pbp.data_nba.jump_ball import DataJumpBall
from pbpstats.resources.enhanced_pbp.data_nba.rebound import DataRebound
from pbpstats.resources.enhanced_pbp.data_nba.replay import DataReplay
from pbpstats.resources.enhanced_pbp.data_nba.start_of_period import DataStartOfPeriod
from pbpstats.resources.enhanced_pbp.data_nba.substitution import DataSubstitution
from pbpstats.resources.enhanced_pbp.data_nba.timeout import DataTimeout
from pbpstats.resources.enhanced_pbp.data_nba.turnover import DataTurnover
from pbpstats.resources.enhanced_pbp.data_nba.violation import DataViolation

__all__ = [
    "DataEjection",
    "DataEndOfPeriod",
    "DataFieldGoal",
    "DataFoul",
    "DataFreeThrow",
    "DataJumpBall",
    "DataRebound",
    "DataReplay",
    "DataStartOfPeriod",
    "DataSubstitution",
    "DataTimeout",
    "DataTurnover",
    "DataViolation",
]



"""
``StatsNbaEnhancedPbpLoader`` loads pbp data for a game and
creates :obj:`~pbpstats.resources.enhanced_pbp.enhanced_pbp_item.EnhancedPbpItem` objects
for each event

Enhanced data for each event includes current players on floor, score, fouls to give and number of fouls committed by each player,
plus additional data depending on event type

The following code will load pbp data for game id "0021900001" from a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import StatsNbaEnhancedPbpFileLoader, StatsNbaEnhancedPbpLoader

    source_loader = StatsNbaEnhancedPbpFileLoader("/data")
    pbp_loader = StatsNbaEnhancedPbpLoader("0021900001", source_loader)
    print(pbp_loader.items[0].data)  # prints dict with the first event of the game
"""
import os
import json

from pbpstats.data_loader.data_nba.pbp.loader import DataNbaPbpLoader
from pbpstats.data_loader.data_nba.pbp.web import DataNbaPbpWebLoader
from pbpstats.data_loader.stats_nba.pbp.loader import StatsNbaPbpLoader
from pbpstats.data_loader.stats_nba.shots.loader import StatsNbaShotsLoader
from pbpstats.data_loader.nba_enhanced_pbp_loader import NbaEnhancedPbpLoader
from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_factory import (
    StatsNbaEnhancedPbpFactory,
)
from pbpstats.resources.enhanced_pbp import FieldGoal
from pbpstats.resources.enhanced_pbp.rebound import EventOrderError


class StatsNbaEnhancedPbpLoader(StatsNbaPbpLoader, NbaEnhancedPbpLoader):
    """
    Loads stats.nba.com source enhanced pbp data for game.
    Events are stored in items attribute as :obj:`~pbpstats.resources.enhanced_pbp.enhanced_pbp_item.EnhancedPbpItem` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.stats_nba.enhanced_pbp.file.StatsNbaEnhancedPbpFileLoader` or :obj:`~pbpstats.data_loader.stats_nba.enhanced_pbp.file.StatsNbaEnhancedPbpWebLoader` object
    :raises: :obj:`~pbpstats.resources.enhanced_pbp.start_of_period.InvalidNumberOfStartersException`:
        If all 5 players that start the period for a team can't be determined.
        You can add the correct period starters to overrides/missing_period_starters.json in your data directory to fix this.
    :raises: :obj:`~pbpstats.resources.enhanced_pbp.rebound.EventOrderError`:
        If rebound event is not immediately following a missed shot event.
        You can manually edit the event order in the pbp file stored on disk to fix this.
    """

    data_provider = "stats_nba"
    resource = "EnhancedPbp"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.shots_source_loader = source_loader.shots_source_loader
        super().__init__(game_id, source_loader)

    def _make_pbp_items(self):
        self.factory = StatsNbaEnhancedPbpFactory()
        self.items = [
            self.factory.get_event_class(item["EVENTMSGTYPE"])(item, i)
            for i, item in enumerate(self.data)
        ]
        self._add_extra_attrs_to_all_events()
        self._check_rebound_event_order(6)
        self._add_shot_x_y_coords()

    def _add_shot_x_y_coords(self):
        shots_loader = StatsNbaShotsLoader(self.game_id, self.shots_source_loader)
        shots_event_num_map = {
            item.game_event_id: {"loc_x": item.loc_x, "loc_y": item.loc_y}
            for item in shots_loader.items
        }
        for item in self.items:
            if (
                isinstance(item, FieldGoal)
                and item.event_num in shots_event_num_map.keys()
            ):
                item.locX = shots_event_num_map[item.event_num]["loc_x"]
                item.locY = shots_event_num_map[item.event_num]["loc_y"]

    def _check_rebound_event_order(self, max_retries):
        """
        checks rebound events to make sure they are ordered correctly
        """
        attempts = 0
        while attempts <= max_retries:
            try:
                for event in self.items:
                    if hasattr(event, "missed_shot"):
                        event.missed_shot
                break
            except EventOrderError as e:
                self._fix_common_event_order_error(e)
                self.items = [
                    self.factory.get_event_class(item["EVENTMSGTYPE"])(item, i)
                    for i, item in enumerate(self.data)
                ]
                self._add_extra_attrs_to_all_events()
                attempts += 1
        # Common check didn't fix things. Use data nba event order
        try:
            for event in self.items:
                if hasattr(event, "missed_shot"):
                    event.missed_shot
        except EventOrderError as e:
            self._use_data_nba_event_order()
            self.items = [
                self.factory.get_event_class(item["EVENTMSGTYPE"])(item, i)
                for i, item in enumerate(self.data)
            ]
            self._add_extra_attrs_to_all_events()

    def _fix_common_event_order_error(self, exception):
        """
        fixs common cases where events are out of order
        current cases are:
        - subs/timeouts between free throws being between missed second FT and rebound
        - end of period replay events being between missed shot and rebound
        - rebound and shot in reverse order
        - shot, rebound, rebound - first rebound needs to me moved to before shot
        - shot, rebound, rebound - second rebound needs to be moved ahead of shot and first rebound
        """
        event_num = int(str(exception).split("EventNum: ")[-1].split(">")[0])
        headers = self.source_data["resultSets"][0]["headers"]
        rows = self.source_data["resultSets"][0]["rowSet"]
        event_num_index = headers.index("EVENTNUM")
        event_type_index = headers.index("EVENTMSGTYPE")
        for i, row in enumerate(rows):
            if row[event_num_index] == int(event_num):
                issue_event_index = i

        if rows[issue_event_index][event_type_index] in [8, 9]:
            # these are subs/timeouts that are in between missed ft and rebound
            # move all sub/timeout events between ft and rebound to before ft
            row_index = issue_event_index
            while rows[row_index][event_type_index] in [8, 9]:
                row_index -= 1

            # rows[row_index] should be moved to right before rows[issue_event_index]
            new_rows = []
            row_to_move_event_num = rows[row_index][event_num_index]
            for row in rows:
                if row[event_num_index] == row_to_move_event_num:
                    row_to_move = row
                elif row[event_num_index] == int(event_num):
                    new_rows.append(row)
                    new_rows.append(row_to_move)
                else:
                    new_rows.append(row)

            self.source_data["resultSets"][0]["rowSet"] = new_rows
        elif (
            rows[issue_event_index][event_type_index] == 18
            and rows[issue_event_index + 1][event_type_index] == 4
        ):
            # these are replays that are in between missed shot and rebound
            # move replay event after rebound
            replay_event = rows[issue_event_index]
            rebound_event = rows[issue_event_index + 1]

            self.source_data["resultSets"][0]["rowSet"][
                issue_event_index + 1
            ] = replay_event
            self.source_data["resultSets"][0]["rowSet"][
                issue_event_index
            ] = rebound_event

        elif (
            rows[issue_event_index + 1][event_type_index] == 4
            and rows[issue_event_index + 1][event_num_index] == int(event_num) - 1
        ):
            # rebound and shot need to be flipped
            rebound_event = rows[issue_event_index + 1]
            shot_event = rows[issue_event_index]
            self.source_data["resultSets"][0]["rowSet"][
                issue_event_index
            ] = rebound_event
            self.source_data["resultSets"][0]["rowSet"][
                issue_event_index + 1
            ] = shot_event
        elif (
            rows[issue_event_index + 1][event_type_index] == 4
            and rows[issue_event_index + 1][event_num_index] == int(event_num) + 2
            and rows[issue_event_index - 1][event_type_index] == 2
            and rows[issue_event_index - 1][event_num_index] == int(event_num) + 1
        ):
            # shot, rebound, rebound - first rebound need to me moved to before shot
            rebound_event = rows[issue_event_index]
            shot_event = rows[issue_event_index - 1]
            self.source_data["resultSets"][0]["rowSet"][issue_event_index] = shot_event
            self.source_data["resultSets"][0]["rowSet"][
                issue_event_index - 1
            ] = rebound_event
        elif (
            rows[issue_event_index + 1][event_type_index] == 4
            and rows[issue_event_index + 1][event_num_index] == int(event_num) - 2
            and rows[issue_event_index - 1][event_type_index] == 2
            and rows[issue_event_index - 1][event_num_index] == int(event_num) - 1
        ):
            # shot, rebound, rebound - second rebound needs to be moved ahead of shot and first rebound
            first_rebound = rows[issue_event_index + 1]
            second_rebound = rows[issue_event_index]
            shot_event = rows[issue_event_index - 1]
            self.source_data["resultSets"][0]["rowSet"][
                issue_event_index + 1
            ] = second_rebound
            self.source_data["resultSets"][0]["rowSet"][issue_event_index] = shot_event
            self.source_data["resultSets"][0]["rowSet"][
                issue_event_index - 1
            ] = first_rebound

        self._save_data_to_file()

    def _use_data_nba_event_order(self):
        """
        reorders all events to be the same order as data.nba.com pbp
        """
        # Order event numbers of events in data.nba.com pbp
        data_nba_pbp = DataNbaPbpLoader(self.game_id, DataNbaPbpWebLoader())
        data_nba_event_num_order = [item.evt for item in data_nba_pbp.items]

        headers = self.source_data["resultSets"][0]["headers"]
        rows = self.source_data["resultSets"][0]["rowSet"]
        event_num_index = headers.index("EVENTNUM")

        # reorder stats.nba.com events to be in same order as data.nba.com events
        new_event_order = []
        for event_num in data_nba_event_num_order:
            for row in rows:
                if row[event_num_index] == event_num:
                    new_event_order.append(row)
        self.source_data["resultSets"][0]["rowSet"] = new_event_order
        self._save_data_to_file()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f"{self.file_directory}/pbp/stats_{self.game_id}.json"
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import (
    StartOfPeriod,
    InvalidNumberOfStartersException,
)


class DataStartOfPeriod(StartOfPeriod, DataEnhancedPbpItem):
    """
    Class for start of period events
    """

    event_type = 12

    def __init__(self, *args):
        super().__init__(*args)

    def get_period_starters(self, file_directory=None):
        """
        Gets player ids of players who started the period for each team
         If players can't be determined from parsing pbp, will try to
        find them by making API request to stats.nba.com boxscore filtered by time.

        :param str file_directory: directory in which overrides subdirectory exists
            containing period starter overrides when period starters can't be determined
            from parsing pbp events
        :returns: dict with list of player ids for each team
            with players on the floor at start of period
        :raises: :obj:`~pbpstats.resources.enhanced_pbp.start_of_period.InvalidNumberOfStartersException`:
            If all 5 players that start the period for a team can't be determined.
        """
        try:
            return self._get_period_starters_from_period_events(file_directory)
        except InvalidNumberOfStartersException:
            # get starters from stats.nba.com boxcore data by time range
            return self._get_starters_from_boxscore_request()



"""
Instantiating a ``Day`` object will load all resources for the ``Day``
object that were set in the settings when the client was instantiated

The following code will instantiate the client and get game data
for games played on 02/03/2020

.. code-block:: python

    from pbpstats.client import Client

    settings = {
        "dir": "/response_data",
        "Games": {"source": "web", "data_provider": "stats_nba"}
    }
    client = Client(settings)
    day = client.Day("02/03/2020", "nba")
    for game in day.games.items:
        print(game.data)
"""
import inspect

import pbpstats.client as client


class Day(object):
    """
    Class for loading resource data from data loaders with a ``parent_object`` of ``Day``

    :param str date: Formatted as MM/DD/YYYY
    :param str league: Options are 'nba', 'wnba' or 'gleague'
    """

    def __init__(self, date, league):
        """
        date format - 'MM/DD/YYYY
        """
        self.date = date
        self.league = league
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        data_loaders = [
            a for a in attributes if a[0].endswith(client.DATA_LOADER_SUFFIX)
        ]
        data_source_map = {
            a[0].replace(client.DATA_SOURCE_SUFFIX, ""): a[1]
            for a in attributes
            if a[0].endswith(client.DATA_SOURCE_SUFFIX)
        }
        for data_loader in data_loaders:
            attr_name = data_loader[0].replace(client.DATA_LOADER_SUFFIX, "")
            source_loader_cls = data_source_map[attr_name]
            source_loader = source_loader_cls(self.data_directory)
            data = data_loader[1](date, league, source_loader)
            resource_cls = getattr(self, attr_name)
            setattr(
                self,
                client.PATTERN.sub("_", attr_name).lower(),
                resource_cls(data.items),
            )



import json
import os

from pbpstats import (
    NBA_STRING,
    G_LEAGUE_STRING,
    D_LEAGUE_STRING,
    WNBA_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)

from pbpstats.data_loader.data_nba.web_loader import DataNbaWebLoader


class DataNbaScheduleWebLoader(DataNbaWebLoader):
    """
     A ``DataNbaScheduleWebLoader`` object should be instantiated and passed into ``DataNbaSchedulLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `data_<league>_<season_year>.json` in the `/schedule` subdirectory.
        If not provided response data will not be saved on disk.
    """

    def __init__(self, file_directory=None):
        self.file_directory = file_directory

    def load_data(self, league, season):
        self.league_string = league
        self.season_year = season.split("-")[0]
        league_url_part = (
            NBA_STRING if self.league_string == G_LEAGUE_STRING else self.league_string
        )
        league_part = (
            D_LEAGUE_STRING
            if self.league_string == G_LEAGUE_STRING
            else self.league_string
        )
        self.url = f"https://data.{league_url_part}.com/data/10s/v2015/json/mobile_teams/{league_part}/{self.season_year}/league/{self.league_id}_full_schedule.json"
        return self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f"{self.file_directory}/schedule/data_{self.league_string}_{self.season_year}.json"
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)

    @property
    def league_id(self):
        """
        Returns League Id for league.

        00 for nba, 10 for wnba, 20 for g-league
        """
        if self.league_string == NBA_STRING:
            return NBA_GAME_ID_PREFIX
        elif self.league_string == WNBA_STRING:
            return WNBA_GAME_ID_PREFIX
        elif self.league_string == G_LEAGUE_STRING:
            return G_LEAGUE_GAME_ID_PREFIX



from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader


class StatsNbaSummaryFileLoader(StatsNbaFileLoader):
    """
    A ``StatsNbaSummaryFileLoader`` object should be instantiated and passed into ``StatsNbaSummaryLoader`` when loading data from file

    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `stats_summary_<game_id>.json` in the `/game_details` subdirectory.
    """

    def __init__(self, file_directory):
        self.file_directory = file_directory

    @check_file_directory
    def load_data(self, game_id):
        self.file_path = (
            f"{self.file_directory}/game_details/stats_summary_{game_id}.json"
        )
        return self._load_data_from_file()



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import Turnover


class LiveTurnover(Turnover, LiveEnhancedPbpItem):
    """
    Class for Turnover events
    """

    action_type = "turnover"

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def is_no_turnover(self):
        return not hasattr(self, "sub_type")

    @property
    def is_bad_pass(self):
        return (
            hasattr(self, "sub_type")
            and self.stripped_sub_type == "badpass"
            and self.is_steal
        )

    @property
    def is_lost_ball(self):
        return (
            hasattr(self, "sub_type")
            and self.stripped_sub_type == "lostball"
            and self.is_steal
        )

    @property
    def is_travel(self):
        return hasattr(self, "sub_type") and self.stripped_sub_type == "traveling"

    @property
    def is_3_second_violation(self):
        return hasattr(self, "sub_type") and self.stripped_sub_type == "3secviolation"

    @property
    def is_shot_clock_violation(self):
        return hasattr(self, "sub_type") and self.stripped_sub_type == "shotclock"

    @property
    def is_offensive_goaltending(self):
        return (
            hasattr(self, "sub_type")
            and self.stripped_sub_type == "offensivegoaltending"
        )

    @property
    def is_lane_violation(self):
        return hasattr(self, "sub_type") and self.stripped_sub_type == "LaneViolation"

    @property
    def is_kicked_ball(self):
        return hasattr(self, "sub_type") and self.stripped_sub_type == "kickedball"

    @property
    def is_step_out_of_bounds(self):
        return (
            hasattr(self, "sub_type")
            and self.stripped_sub_type == "outofbounds"
            and self.descriptor == "step"
        )

    @property
    def is_lost_ball_out_of_bounds(self):
        return (
            hasattr(self, "sub_type")
            and self.stripped_sub_type == "outofbounds"
            and self.descriptor == "lost ball"
        )

    @property
    def is_bad_pass_out_of_bounds(self):
        return (
            hasattr(self, "sub_type")
            and self.stripped_sub_type == "outofbounds"
            and self.descriptor == "bad pass"
        )



import json
import os

from pbpstats import NBA_STRING, D_LEAGUE_STRING
from pbpstats.data_loader.data_nba.web_loader import DataNbaWebLoader


class DataNbaPbpWebLoader(DataNbaWebLoader):
    """
    A ``DataNbaPbpWebLoader`` object should be instantiated and passed into ``DataNbaPbpLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `data_<game_id>.json` in the `/pbp` subdirectory.
        If not provided response data will not be saved on disk.
    """

    def __init__(self, file_directory=None):
        self.file_directory = file_directory

    def load_data(self, game_id):
        self.game_id = game_id
        league_url_part = NBA_STRING if self.league == D_LEAGUE_STRING else self.league
        self.url = f"https://data.{league_url_part}.com/data/v2015/json/mobile_teams/{self.league}/{self.season}/scores/pbp/{self.game_id}_full_pbp.json"
        return self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f"{self.file_directory}/pbp/data_{self.game_id}.json"
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)



from pbpstats.data_loader.data_nba.boxscore.file import DataNbaBoxscoreFileLoader
from pbpstats.data_loader.data_nba.boxscore.loader import DataNbaBoxscoreLoader
from pbpstats.data_loader.data_nba.boxscore.web import DataNbaBoxscoreWebLoader
from pbpstats.data_loader.data_nba.enhanced_pbp.file import DataNbaEnhancedPbpFileLoader
from pbpstats.data_loader.data_nba.enhanced_pbp.loader import DataNbaEnhancedPbpLoader
from pbpstats.data_loader.data_nba.enhanced_pbp.web import DataNbaEnhancedPbpWebLoader
from pbpstats.data_loader.data_nba.pbp.file import DataNbaPbpFileLoader
from pbpstats.data_loader.data_nba.pbp.loader import DataNbaPbpLoader
from pbpstats.data_loader.data_nba.pbp.web import DataNbaPbpWebLoader
from pbpstats.data_loader.data_nba.possessions.file import DataNbaPossessionFileLoader
from pbpstats.data_loader.data_nba.possessions.loader import DataNbaPossessionLoader
from pbpstats.data_loader.data_nba.possessions.web import DataNbaPossessionWebLoader
from pbpstats.data_loader.data_nba.schedule.file import DataNbaScheduleFileLoader
from pbpstats.data_loader.data_nba.schedule.loader import DataNbaScheduleLoader
from pbpstats.data_loader.data_nba.schedule.web import DataNbaScheduleWebLoader
from pbpstats.data_loader.live.boxscore.file import LiveBoxscoreFileLoader
from pbpstats.data_loader.live.boxscore.loader import LiveBoxscoreLoader
from pbpstats.data_loader.live.boxscore.web import LiveBoxscoreWebLoader
from pbpstats.data_loader.live.enhanced_pbp.file import LiveEnhancedPbpFileLoader
from pbpstats.data_loader.live.enhanced_pbp.loader import LiveEnhancedPbpLoader
from pbpstats.data_loader.live.enhanced_pbp.web import LiveEnhancedPbpWebLoader
from pbpstats.data_loader.live.pbp.file import LivePbpFileLoader
from pbpstats.data_loader.live.pbp.loader import LivePbpLoader
from pbpstats.data_loader.live.pbp.web import LivePbpWebLoader
from pbpstats.data_loader.live.possessions.file import LivePossessionFileLoader
from pbpstats.data_loader.live.possessions.loader import LivePossessionLoader
from pbpstats.data_loader.live.possessions.web import LivePossessionWebLoader
from pbpstats.data_loader.stats_nba.boxscore.file import StatsNbaBoxscoreFileLoader
from pbpstats.data_loader.stats_nba.boxscore.loader import StatsNbaBoxscoreLoader
from pbpstats.data_loader.stats_nba.boxscore.web import StatsNbaBoxscoreWebLoader
from pbpstats.data_loader.stats_nba.enhanced_pbp.file import (
    StatsNbaEnhancedPbpFileLoader,
)
from pbpstats.data_loader.stats_nba.enhanced_pbp.loader import StatsNbaEnhancedPbpLoader
from pbpstats.data_loader.stats_nba.enhanced_pbp.web import StatsNbaEnhancedPbpWebLoader
from pbpstats.data_loader.stats_nba.game_finder.file import StatsNbaGameFinderFileLoader
from pbpstats.data_loader.stats_nba.game_finder.loader import StatsNbaGameFinderLoader
from pbpstats.data_loader.stats_nba.game_finder.web import StatsNbaGameFinderWebLoader
from pbpstats.data_loader.stats_nba.pbp.file import StatsNbaPbpFileLoader
from pbpstats.data_loader.stats_nba.pbp.loader import StatsNbaPbpLoader
from pbpstats.data_loader.stats_nba.pbp.web import StatsNbaPbpWebLoader
from pbpstats.data_loader.stats_nba.possessions.file import StatsNbaPossessionFileLoader
from pbpstats.data_loader.stats_nba.possessions.loader import StatsNbaPossessionLoader
from pbpstats.data_loader.stats_nba.possessions.web import StatsNbaPossessionWebLoader
from pbpstats.data_loader.stats_nba.scoreboard.file import StatsNbaScoreboardFileLoader
from pbpstats.data_loader.stats_nba.scoreboard.loader import StatsNbaScoreboardLoader
from pbpstats.data_loader.stats_nba.scoreboard.web import StatsNbaScoreboardWebLoader
from pbpstats.data_loader.stats_nba.shots.file import StatsNbaShotsFileLoader
from pbpstats.data_loader.stats_nba.shots.loader import StatsNbaShotsLoader
from pbpstats.data_loader.stats_nba.shots.web import StatsNbaShotsWebLoader

__all__ = [
    "DataNbaBoxscoreFileLoader",
    "DataNbaBoxscoreLoader",
    "DataNbaBoxscoreWebLoader",
    "DataNbaEnhancedPbpFileLoader",
    "DataNbaEnhancedPbpLoader",
    "DataNbaEnhancedPbpWebLoader",
    "DataNbaPbpFileLoader",
    "DataNbaPbpLoader",
    "DataNbaPbpWebLoader",
    "DataNbaPossessionFileLoader",
    "DataNbaPossessionLoader",
    "DataNbaPossessionWebLoader",
    "DataNbaScheduleFileLoader",
    "DataNbaScheduleLoader",
    "DataNbaScheduleWebLoader",
    "LiveBoxscoreFileLoader",
    "LiveBoxscoreLoader",
    "LiveBoxscoreWebLoader",
    "LiveEnhancedPbpFileLoader",
    "LiveEnhancedPbpLoader",
    "LiveEnhancedPbpWebLoader",
    "LivePbpFileLoader",
    "LivePbpLoader",
    "LivePbpWebLoader",
    "LivePossessionFileLoader",
    "LivePossessionLoader",
    "LivePossessionWebLoader",
    "StatsNbaBoxscoreFileLoader",
    "StatsNbaBoxscoreLoader",
    "StatsNbaBoxscoreWebLoader",
    "StatsNbaEnhancedPbpFileLoader",
    "StatsNbaEnhancedPbpLoader",
    "StatsNbaEnhancedPbpWebLoader",
    "StatsNbaGameFinderFileLoader",
    "StatsNbaGameFinderLoader",
    "StatsNbaGameFinderWebLoader",
    "StatsNbaPbpFileLoader",
    "StatsNbaPbpLoader",
    "StatsNbaPbpWebLoader",
    "StatsNbaPossessionFileLoader",
    "StatsNbaPossessionLoader",
    "StatsNbaPossessionWebLoader",
    "StatsNbaScoreboardFileLoader",
    "StatsNbaScoreboardLoader",
    "StatsNbaScoreboardWebLoader",
    "StatsNbaShotsFileLoader",
    "StatsNbaShotsLoader",
    "StatsNbaShotsWebLoader",
]



"""
``DataNbaEnhancedPbpFactory`` can be used to create enhanced pbp event objects from
the :mod:`pbpstats.resources.enhanced_pbp.data_nba` module based on the event type.

The following code will get the event class for event type 1 (which is the event type for a field goal make)

.. code-block:: python

    from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_factory import DataNbaEnhancedPbpFactory

    factory = DataNbaEnhancedPbpFactory()
    event_class = factory.get_event_class(1)
    print(event_class)  # prints "<class 'pbpstats.resources.enhanced_pbp.data_nba.field_goal.DataFieldGoal'>"
"""
import pbpstats.resources.enhanced_pbp.data_nba as event_types
from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)


class DataNbaEnhancedPbpFactory(object):
    """
    Class for factory of event type classes. On initialization will load
    in all event classes in the :mod:`pbpstats.resources.enhanced_pbp.data_nba` module
    """

    def __init__(self):
        self.event_classes = {}
        self._load_event_classes()

    def _load_event_classes(self):
        event_classes = dict(
            [
                (name, cls)
                for name, cls in event_types.__dict__.items()
                if isinstance(cls, type)
            ]
        )
        for _, event_cls in event_classes.items():
            if isinstance(event_cls.event_type, list):
                for event_type in event_cls.event_type:
                    self.event_classes[event_type] = event_cls
            else:
                self.event_classes[event_cls.event_type] = event_cls

    def get_event_class(self, event_type):
        """
        Gets the class for the event based on the event_type

        :param int event_type: event action type for the event

        :returns: class for event type
        """
        return self.event_classes.get(event_type, DataEnhancedPbpItem)



class EndOfPeriod(object):
    """
    Class for end of period events
    """

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        return self.base_stats



from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.data_nba.file_loader import DataNbaFileLoader


class DataNbaScheduleFileLoader(DataNbaFileLoader):
    """
    A ``DataNbaScheduleFileLoader`` object should be instantiated and passed into ``DataNbaScheduleLoader`` when loading data from file

    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `data_<league>_<season_year>.json` in the `/schedule` subdirectory.
    """

    def __init__(self, file_directory=None):
        self.file_directory = file_directory

    @check_file_directory
    def load_data(self, league, season):
        self.league_string = league
        self.season_year = season.split("-")[0]
        self.file_path = f"{self.file_directory}/schedule/data_{self.league_string}_{self.season_year}.json"
        return self._load_data_from_file()



import abc

import pbpstats
from pbpstats.resources.enhanced_pbp import (
    EndOfPeriod,
    FieldGoal,
    FreeThrow,
    Replay,
    Turnover,
)


class EventOrderError(Exception):
    """
    Class for exception raised when rebound event is
    not immediately following a missed shot event.

    You can manually edit the event order in the pbp file
    stored on disk to fix this.
    """

    pass


class Rebound(object):
    """
    Class for rebound events
    """

    @property
    def is_real_rebound(self):
        """
        Returns True if rebound should be counted as a rebound, False otherwise.

        All missed shots have a rebound in the play-by-play but
        not all of these rebounds should be counted as actual rebounds. Some are just
        placeholder events.
        """
        if self.is_placeholder:
            return False

        if self.is_buzzer_beater_placeholder:
            return False

        if self.is_turnover_placeholder:
            return False

        if self.is_non_live_ft_placeholder:
            return False

        if self.is_buzzer_beater_rebound_at_shot_time:
            return False

        return True

    @abc.abstractproperty
    def is_placeholder(self):
        """
        returns True if rebound is a placeholder event, False otherwise.

        These are team rebounds on for example missed FT 1 of 2
        """
        return self.event_action_type != 0 and self.player1_id == 0

    @property
    def is_turnover_placeholder(self):
        """
        returns True if rebound is a placeholder event when a turnover occurs, False otherwise.

        Example shot clock violation or kicked ball turnover at time of team rebound
        """
        events_at_event_time = self.get_all_events_at_current_time()
        for event in events_at_event_time:
            if (
                isinstance(event, Turnover)
                and (event.is_shot_clock_violation or event.is_kicked_ball)
            ) and self.player1_id == 0:
                return True
        return False

    @property
    def is_non_live_ft_placeholder(self):
        """
        returns True if rebound is a placeholder event after a missed free throw that is not a live ball, False otherwise.

        Example: rebound after missed flagrant FT 2 of 2
        """
        if isinstance(self.missed_shot, FreeThrow) and not self.missed_shot.is_end_ft:
            return True
        return False

    @property
    def is_buzzer_beater_placeholder(self):
        """
        returns True if rebound is a placeholder event after a missed buzzer beater, False otherwise.

        Rebounds occur after time has expired but are still logged in play-by-play,
        but should not be counted in rebound totals
        """
        next_event = self.next_event
        if isinstance(next_event, Replay):
            next_event = next_event.next_event
        if (
            (self.clock == "00:00.0" or self.clock == "0:00")
            and self.player1_id == 0
            and (
                next_event is not None
                and isinstance(next_event, EndOfPeriod)
                or next_event is None
            )
        ):
            return True
        return False

    @property
    def is_buzzer_beater_rebound_at_shot_time(self):
        """
        returns True if rebound is a placeholder event after a missed buzzer beater, False otherwise.

        Sometimes rebound on buzzer beater is given the same time as shot - don't count these.
        Only don't count if rebound is last event before end of period event, ignoring replay events
        """
        if (
            self.missed_shot.seconds_remaining <= 3
            and self.seconds_remaining == self.missed_shot.seconds_remaining
            and self.player1_id == 0
        ):
            next_event = self.next_event
            if isinstance(next_event, Replay):
                next_event = next_event.next_event
            if isinstance(next_event, EndOfPeriod):
                return True
        return False

    @abc.abstractproperty
    def missed_shot(self):
        pass

    @abc.abstractproperty
    def oreb(self):
        pass

    @property
    def self_reb(self):
        """
        returns True if rebound was gotten by player who missed the shot, False otherwise
        """
        return self.player1_id == self.missed_shot.player1_id

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        stats = []
        if self.is_real_rebound:
            shot_type = self.missed_shot.shot_type
            if isinstance(self.missed_shot, FieldGoal) and self.missed_shot.is_blocked:
                if not self.oreb:
                    blocked_recovered_key = (
                        pbpstats.BLOCKED_STRING + shot_type + "Recovered"
                    )
                    block_player_id = self.missed_shot.player3_id
                    stats.append(
                        {
                            "player_id": block_player_id,
                            "team_id": self.team_id,
                            "stat_key": blocked_recovered_key,
                            "stat_value": 1,
                        }
                    )
                shot_type += pbpstats.BLOCKED_STRING
            if self.oreb:
                reb_key = (
                    shot_type
                    + pbpstats.OFFENSIVE_ABBREVIATION_PREFIX
                    + pbpstats.REBOUNDS_STRING
                )
                opportunity_key = (
                    shot_type
                    + pbpstats.OFFENSIVE_ABBREVIATION_PREFIX
                    + pbpstats.REBOUND_OPPORTUNITIES_STRING
                )
                opponent_opportunity_key = (
                    shot_type
                    + pbpstats.DEFENSIVE_ABBREVIATION_PREFIX
                    + pbpstats.REBOUND_OPPORTUNITIES_STRING
                )
            else:
                reb_key = (
                    shot_type
                    + pbpstats.DEFENSIVE_ABBREVIATION_PREFIX
                    + pbpstats.REBOUNDS_STRING
                )
                opportunity_key = (
                    shot_type
                    + pbpstats.DEFENSIVE_ABBREVIATION_PREFIX
                    + pbpstats.REBOUND_OPPORTUNITIES_STRING
                )
                opponent_opportunity_key = (
                    shot_type
                    + pbpstats.OFFENSIVE_ABBREVIATION_PREFIX
                    + pbpstats.REBOUND_OPPORTUNITIES_STRING
                )

            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": reb_key,
                    "stat_value": 1,
                }
            )

            # rebound opportunites for all players on floor
            for team_id, players in self.current_players.items():
                stat_key = (
                    opportunity_key
                    if team_id == self.team_id
                    else opponent_opportunity_key
                )
                for player_id in players:
                    stat_item = {
                        "player_id": player_id,
                        "team_id": team_id,
                        "stat_key": stat_key,
                        "stat_value": 1,
                    }
                    stats.append(stat_item)
                if team_id == self.team_id and self.oreb:
                    for player_id in players:
                        on_floor_oreb_stat_item = {
                            "player_id": player_id,
                            "team_id": team_id,
                            "stat_key": pbpstats.ON_FLOOR_OFFENSIVE_REBOUND_STRING,
                            "stat_value": 1,
                        }
                        stats.append(on_floor_oreb_stat_item)

            # player missed shot rebound stats
            shooter_player_id = self.missed_shot.player1_id
            missed_shot_rebounded_opportunity_key = (
                shot_type
                + pbpstats.OFFENSIVE_ABBREVIATION_PREFIX
                + pbpstats.REBOUNDED_OPPORTUNITIES_STRING
            )
            stats.append(
                {
                    "player_id": shooter_player_id,
                    "team_id": self.missed_shot.team_id,
                    "stat_key": missed_shot_rebounded_opportunity_key,
                    "stat_value": 1,
                }
            )
            if self.oreb:
                missed_shot_rebounded_key = (
                    shot_type
                    + pbpstats.OFFENSIVE_ABBREVIATION_PREFIX
                    + pbpstats.REBOUNDED_STRING
                )
                stats.append(
                    {
                        "player_id": shooter_player_id,
                        "team_id": self.missed_shot.team_id,
                        "stat_key": missed_shot_rebounded_key,
                        "stat_value": 1,
                    }
                )
                if self.self_reb:
                    self_rebounded_key = shot_type + pbpstats.SELF_REBOUND_STRING
                    stats.append(
                        {
                            "player_id": shooter_player_id,
                            "team_id": self.missed_shot.team_id,
                            "stat_key": self_rebounded_key,
                            "stat_value": 1,
                        }
                    )

            team_ids = list(self.current_players.keys())
            opponent_team_id = (
                team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
            )
            lineups_ids = self.lineup_ids
            for stat in stats:
                opponent_team_id = (
                    team_ids[0] if stat["team_id"] == team_ids[1] else team_ids[1]
                )
                stat["lineup_id"] = lineups_ids[stat["team_id"]]
                stat["opponent_team_id"] = opponent_team_id
                stat["opponent_lineup_id"] = lineups_ids[opponent_team_id]
        return self.base_stats + stats



import json
from pathlib import Path

from pbpstats.data_loader.data_nba.base import DataNbaLoaderBase


class DataNbaFileLoader(DataNbaLoaderBase):
    """
    Base class for loading data.nba.com files saved on disk.

    All data.nba.com data loader classes should inherit from this class.

    This class should not be instantiated directly.
    """

    def _load_data_from_file(self):
        data_file = Path(self.file_path)
        if not data_file.is_file():
            raise Exception(f"{self.file_path} does not exist")
        with open(self.file_path) as json_data:
            self.source_data = json.load(json_data)
            return self.source_data
